/*
 * Copyright (c) 1999 The Java Apache Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software and design ideas developed by the Java 
 *    Apache Project (http://java.apache.org/)."
 *
 * 4. The names "Cocoon", "Cocoon Servlet" and "Java Apache Project" must 
 *    not be used to endorse or promote products derived from this software 
 *    without prior written permission.
 *
 * 5. Products derived from this software may not be called "Cocoon"
 *    nor may "Cocoon" and "Java Apache Project" appear in their names without 
 *    prior written permission of the Java Apache Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software and design ideas developed by the Java 
 *    Apache Project (http://java.apache.org/)."
 *           
 * THIS SOFTWARE IS PROVIDED BY THE JAVA APACHE PROJECT "AS IS" AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE JAVA APACHE PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *           
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Java Apache Project. For more information
 * on the Java Apache Project please see <http://java.apache.org/>.
 */
 
package org.apache.cocoon;

import java.io.*;
import java.net.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import org.w3c.dom.*;
import org.apache.cocoon.cache.*;
import org.apache.cocoon.parser.*;
import org.apache.cocoon.printer.*;
import org.apache.cocoon.processor.*;
import org.apache.cocoon.framework.*;

/**
 * The Cocoon servlet.
 *
 * This servlet implements an XML/XSL server side publishing framework to
 * separate different knowledge contexts in different processing layers.
 *
 * @author <a href="mailto:stefano@apache.org">Stefano Mazzocchi</a>
 * @author <a href="mailto:ritter@wt.net">Greg Ritter</a>
 * @version $Revision: 1.3 $ $Date: 1999/03/24 13:46:05 $
 */
 
public class Cocoon extends HttpServlet implements CocoonDefaults {
    
    Parser parser;
    Vector processors;
    Printer printer;
    Browsers browsers;
    Configurations configurations;
    Cache cache;

    /**
     * This method initializes the servlet.
     */       
    public void init(ServletConfig config) throws ServletException {
        super.init(config);

        // Collect the configurations
    	try {
            configurations = ConfigurationFactory.getConfigurations(config.getInitParameter(INIT_ARG), INTERNAL_PROPERTIES);
    	} catch (IOException e) {
    		throw new ServletException(e.getMessage() + ".  Properties file is: " + config.getInitParameter(INIT_ARG));
    	}

        // Create the browser table
        browsers = new Browsers(configurations.getConfigurations(BROWSERS));

        // Set the factory configurations
        Factory.init(configurations);
        
        // Create the parser
        parser = (Parser) Factory.create(PARSER);
        
        // Prepare the processor context
        Configurations context = new Configurations();
        context.set("parser", parser);
        
        // Create the processor pipeline
        processors = (Vector) Factory.create(PROCESSOR, context);

        // Create the printer
        printer = (Printer) Factory.create(PRINTER);

        // Prepare the cache context
        context = new Configurations();
        context.set("memsize", configurations.get("cache.memsize"));
        context.set("disksize", configurations.get("cache.disksize"));
        context.set("directory", configurations.get("cache.directory"));

        // Create the cache
        cache = (Cache) Factory.create(CACHE,context);
    	
    	if (DEBUG) {
    		log("Browsers: " + browsers);
    		log("Parser: " + parser);
    		log("Processors: " + processors);
    		log("Printer: " + printer);
            log("Cache: " + cache);
    	}
    }

    /**
     * This method is used to abstract the resource from the file system like
     * mandated by the Servlet API 2.1 but to let this servlet remain
     * compatible with 2.0 system until 2.1 systems are widely available.
     */ 
    public URLConnection getResourceConnection(String path) {
        try {
            return ((new File(path)).toURL()).openConnection();
        } catch (Exception e) {
            return null;
        }
    }        
        
    /**
     * This method is called by the servlet engine to handle the request.
     */ 
    public void service(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {

    	long time = 0;
        boolean fromCache = false;
        String basename = request.getPathTranslated();
        String browser = browsers.mapBrowser(request.getHeader("user-Agent"));

        if (VERBOSE) {
            time = System.currentTimeMillis();
        }
    	
        // set content type and other response header fields first
        // FIXME: is Cocoon going to return HTML alone? Should we buffer
        // the servlet output stream to know which type it is at the end? (SM)
        response.setContentType("text/html");

        // get output stream
        PrintWriter out = new PrintWriter(response.getWriter());

        try {
            // get requested resource as URLConnection
            URLConnection resource = getResourceConnection(basename);

            if (resource != null) {
                // if a cached copy exists for this basename/browser, and
                // the source has not changed since it was cached, use it
                File resourceFile = new File(basename);
            	String fullname = basename + "." + browser;
                if (cache.containsKey(fullname) && resourceFile.lastModified() <= ((CachedObject) cache.get(fullname)).getLastModified()) {
                    out.print(((CachedObject) cache.get(fullname)).getContents());
                    fromCache = true;
                } else {
                    // create the document by parsing the file
                    Document document = parser.parse(new InputStreamReader(resource.getInputStream()));
        
                    // create the processor parameters
                    String file = (new File(resource.getURL().getFile())).getPath();
                    String path = file.substring(0, file.lastIndexOf(File.separator) + 1);
                        
                    // Printout the strings for debug
                    if (DEBUG) {
                        log("File: " + file);
                        log("Path: " + path);
                        log("Browser: " + browser);
                    }
            	
                    // place them into the parameters hashtable
                    Hashtable parameters = new Hashtable();
                    parameters.put("path", path);
                    parameters.put("browser", browser);
        
                    // process the document thru the processor pipeline.
                    Enumeration stages = processors.elements();
                    while (stages.hasMoreElements()) {
                        document = ((Processor) stages.nextElement()).process(document, parameters);
		            }
                
                    // write and format the document tree onto the output stream
                    StringWriter cacheStringWriter = new StringWriter();
                    printer.print(document, new PrintWriter(cacheStringWriter));                
                    out.print(cacheStringWriter.toString());
                    cache.put(fullname, cacheStringWriter.toString());
                }
            } else {
                error("Could not access the given URI <b>" + request.getRequestURI()
                    + "</b><p><b>NOTE</b>: this servlet cannot work if the requested "
                    + "file is not accessible by the servlet engine. "
                    + "Please, make sure the requested file is readable by "
                    + "this servlet and reachable by the executing system.</p>", out);
            }
        } catch (Exception e) {
            StringWriter buffer = new StringWriter();
            e.printStackTrace(new PrintWriter(buffer));
            error (buffer.toString(), out);
        }
        
    	if (VERBOSE) {
            // Evaluate elapsed time
            time = System.currentTimeMillis() - time;
            
            // Print time comment
            if (fromCache) {
                out.println("\n<!-- This page was served from cache by " + getServletInfo() + " in " + time + " milliseconds -->");
            } else {
                out.println("\n<!-- This page was dynamically created and cached by " + getServletInfo() + " in " + time + " milliseconds -->");
            }
            
            // Print cache diagnostics
            out.println("\n<!-- Memory cache contains " + cache.getSize() + " of " + cache.getMaxMemSize() + " objects (up to " + cache.getMaxDiskSize() + " objects may be stored on disk) -->");
    	}
        
        // close output stream
        out.close();
        
    }
    
    private void error(String message, PrintWriter out) throws IOException {
        out.print("<HTML><BODY><CENTER>"
            + "<H2>Servlet Error</H2>"
            + "<P><I>" + message + "</I></P>" 
            + "<HR width=\"50%\">Automatically generated by:<BR>"
            + "<B>" + getServletInfo() + "</B>"
            + "</CENTER></BODY></HTML>");
    }
    
    public String getServletInfo() {
        return NAME + " " + VERSION;
    }
}