/*
 * Copyright (c) 1999 The Java Apache Project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software and design ideas developed by the Java 
 *    Apache Project (http://java.apache.org/)."
 *
 * 4. The names "Cocoon", "Cocoon Servlet" and "Java Apache Project" must 
 *    not be used to endorse or promote products derived from this software 
 *    without prior written permission.
 *
 * 5. Products derived from this software may not be called "Cocoon"
 *    nor may "Cocoon" and "Java Apache Project" appear in their names without 
 *    prior written permission of the Java Apache Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software and design ideas developed by the Java 
 *    Apache Project (http://java.apache.org/)."
 *           
 * THIS SOFTWARE IS PROVIDED BY THE JAVA APACHE PROJECT "AS IS" AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE JAVA APACHE PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *           
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Java Apache Project. For more information
 * on the Java Apache Project please see <http://java.apache.org/>.
 */
 
package org.apache.cocoon.cache;

import java.io.*;
import java.util.*;
import org.apache.cocoon.framework.*;

/**
 * A class to encapsulate the caching of preprocessed XML files.
 *
 * @author <a href="mailto:ritter@wt.net">Greg Ritter</a>
 * @version $Revision: 1.2 $ $Date: 1999/03/24 16:05:56 $
 */
 
public class LRUCache implements Configurable, Cache {
	
    private Hashtable cache;
    private int maxMemObjects = 0;
    private int maxDiskObjects = 0;
    private String directory;

    /**
     * Initialize this cache using the configuration information.
     * @param conf the configuration information
     */
    public void init(Configurations conf) {
    	try {
            this.maxMemObjects = Integer.parseInt(conf.getProperty("memsize"));
    		if (this.maxMemObjects < 0) this.maxMemObjects = 0;
    	} catch (Exception e) {}
    	
        try {
            this.maxDiskObjects = Integer.parseInt(conf.getProperty("disksize"));
            if (this.maxDiskObjects < 0) this.maxDiskObjects = 0;
        } catch (Exception e) {}
    	
        try {
            this.directory = conf.getProperty("directory");
        } catch (Exception e) {}
    	
        if (maxMemObjects > 0) {
            cache = new Hashtable(maxMemObjects);
        } else {
            cache = new Hashtable();
        }
    }

    /**
     * Put a key/value pair in the cache.
     * If necessary, move the least recently used cache pair to disk
     * to make room for the new pair.
     * @param key the key Object of the key/value pair
     * @param value the value Object of the key/value pair
     */
    public synchronized void put(Object key, Object value) {
        if ((maxMemObjects == 0) || (cache.size() < maxMemObjects)) {
            cache.put(key, new CachedObject(value));
        } else {
            putLRUToDisk();
            cache.put(key, new CachedObject(value));
        }
    }

    /**
     * Get a value from the cache for the specified key Object.
     * If necessary, get the value from disk.
     * @param key the key Object
     * @return the value for the key, or null if the key was not in the cache
     */
    public synchronized Object get(Object key) {
        Object temp;
    	
        if ((temp = cache.get(key)) == null) {
            temp = new CachedObject(getFromDisk(key));
            put(key,((CachedObject) temp).getContents());
        }
    	
        return temp;
    }

    /**
     * Check if the specified key exists in this Cache.
     * @param key the key to check for
     * @return true if the key exists in the cache, false otherwise
     */
    public boolean containsKey(Object key) {
        if (cache.containsKey(key)) {
            return true;
        } else {
            return containsKeyOnDisk(key);
        }
    }

    /**
     * Get the number of objects currently in cache.
     * @return the number of objects currently in cache
     */
    public int getSize() {
	    return cache.size() + getSizeOnDisk();
    }

    /**
     * Get the maximum number of objects this cache can hold in memory.
     * @return the number of objects this cache can hold in memory
     */
    public int getMaxMemSize() {
	    return maxMemObjects;
    }

    /**
     * Get the maximum number of objects this cache can hold on disk.
     * @return the number of objects this cache can hold on disk
     */
    public int getMaxDiskSize() {
	    return maxDiskObjects;
    }

    /**
     * Put the least recently used cache pair to disk.
     */
    private void putLRUToDisk() {
        long minTimestamp = -1;
        Object keyWithMinTimestamp = null;
        Enumeration keys = cache.keys();
    	
        while (keys.hasMoreElements()) {
            Object key = keys.nextElement();
            CachedObject value = (CachedObject) cache.get(key);
            if ((minTimestamp == -1) || (minTimestamp > value.getLastAccessed())) {
                minTimestamp = value.getLastAccessed();
                keyWithMinTimestamp = key;
            }
        }
    	
        putToDisk(keyWithMinTimestamp);
    }

    /**
     * Put the object with the specified key to the disk cache.
     * @param key the key of the object to cache to disk
     */
    private void putToDisk(Object key) {
        String keyString = (String) key;
        File path = new File(directory + keyString.substring(0, keyString.replace('\\','/').lastIndexOf('/') + 1));
    	
        if (!path.exists()) {
            path.mkdirs();
        }
    	
        if (path.isDirectory()) {
            File file = new File(directory + keyString);
            try {
                FileWriter fileWriter = new FileWriter(file);
                fileWriter.write(cache.get(key).toString());
                fileWriter.close();
            } catch(IOException ioe) {
                ioe.printStackTrace();
            }
        }
    	
        cache.remove(key);
    }

    /**
     * Get a cached Object from disk for the specified key.
     * @param key the key Object
     * @return the value Object for the specified key
     */
    private Object getFromDisk(Object key) {
        String keyString = (String) key;
        File file = new File(directory + keyString);
        StringBuffer stringBuffer = new StringBuffer();
    	
        try {
            FileReader fileReader = new FileReader(file);
            BufferedReader bufferedReader = new BufferedReader(fileReader);
            String temp;
            while ((temp = bufferedReader.readLine()) != null) {
                stringBuffer.append(temp + "\n");
            }
        } catch(FileNotFoundException fnfe) {
            fnfe.printStackTrace();
        } catch(IOException ioe) {
            ioe.printStackTrace();
        }
    	
        return stringBuffer.toString();
    }

    /**
     * Check if the specified key exists in cache on disk.
     * @param key the key to check for
     * @return true if the key exists on disk, false otherwise
     */
    private boolean containsKeyOnDisk(Object key) {
        File file = new File(directory + (String) key);
        if (file.isFile()) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * Get the number of cached objects on disk for this cache.
     * @return the number of cached objects on disk for this cache
     */
    private int getSizeOnDisk() {
    	return 0;
    }
	
    /**
     * Return a textual identifier for this cache.
     */
    public String toString() {
        return "Cocoon LRU Cache";
    }
}